@@ -21,16 +21,18 @@ class Clusterer(object):
             data: The data to be clustered (default = None).
     """ 
     
-    def __init__(self, nr_clus, x_train=None, y_train=None, data=None, verbose=False):
+    def __init__(self, nr_clus, x_train=None, y_train=None, data=None, relational_data=None, verbose=False):
         self.x_train=x_train
         self.y_train=y_train
         self.nr_clus = nr_clus
         self._verbose = verbose
 
-        if data is None and (x_train is None or y_train is None):
+        if data is None and (x_train is None or y_train is None) and relational_data is None:
             raise Exception("Please specify a valid dataset for clustering.")
         elif data is not None:
         	self.data=data
+        elif relational_data is not None:
+            self.relational_data=relational_data
         else:
             self.data=np.concatenate((self.x_train, np.expand_dims(self.y_train, axis=1)), axis=1)#.reshape(len(self.y_train),1)),axis=1)
 
@@ -57,6 +59,12 @@ class Clusterer(object):
             self.m = kwargs["m"]
         except:
             self.m = 2
+            
+        if method == "fcm" or method == 'gk' or method == 'GK' or method == 'Gustafson-Kessel' or method == 'gustafson-kessel' or method == 'g-k' or method == "pfcm" or method == "fst-pso" or method == "fstpso":
+            assert self.data is not None or (self.x_train is not None and self.y_train is not None), 'Please specify the data tobe clustered.'
+        elif method == 'RFCM' or method == "relational_clustering" or method == "relational" or method == "RC":
+            assert self.relational_data is not None, 'Please specify the relational data matrix.'
+            
 
         if method == "fcm":
             try:
@@ -67,23 +75,19 @@ class Clusterer(object):
                 error = kwargs["fcm_error"]
             except:
                 error = 0.005
-#            try:
-#                seed = kwargs["seed"]
-#            except:
-#                seed = None
+
             centers, partition_matrix, jm = self._fcm(data=self.data, n_clusters=self.nr_clus, m=self.m, max_iter=max_iter, error=error)
         
         elif method == 'gk' or method == 'GK' or method == 'Gustafson-Kessel' or method == 'gustafson-kessel' or method == 'g-k':
             try:
                 max_iter = kwargs["gk_max_iter"]
             except:
-                max_iter=100
+                max_iter=1000
             try:
                 error = kwargs["gk_error"]
             except:
                 error = 0.005
             centers, partition_matrix, jm =self._gk(m=self.m, max_iter=max_iter) 
-            print(partition_matrix)               
 
         elif method == "pfcm":
             try:
@@ -109,7 +113,7 @@ class Clusterer(object):
 
             centers, partition_matrix, typicality_matrix, jm = self._pfcm(data=self.data, n_clusters=self.nr_clus, m=self.m, n=n, max_iter=max_iter, error=error, a=a, b=b)    
 
-        elif method == "fst-pso":
+        elif method == "fst-pso" or method == "fstpso":
             try:
                 max_iter = kwargs["fstpso_max_iter"]
             except:
@@ -126,9 +130,24 @@ class Clusterer(object):
                 path_sol_dump = kwargs["fstpso_path_sol_dump"]
             except:
                 path_sol_dump = None
-
             centers, partition_matrix, jm = self._fstpso(data=self.data, n_clusters=self.nr_clus, max_iter=max_iter, n_particles=n_particles, m=self.m, path_fit_dump=path_fit_dump, path_sol_dump=path_sol_dump)
-
+               
+        elif method == 'RFCM' or method == 'rfcm' or method == "relational_clustering" or method == "relational" or method == "RC":
+            try:
+                max_iter = kwargs["RFCM_max_iter"]
+            except:
+                max_iter=100
+            try:
+                error = kwargs["RFCM_error"]
+            except:
+                error = 0.005
+            try:
+                error = kwargs["RFCM_initialization"]
+            except:
+                initialization = 'random_initialization'
+            
+            centers, partition_matrix, jm = self._rfcm(R=self.relational_data, c=self.nr_clus, m = self.m, epsilon = error, maxIter = max_iter, initType = initialization)            
+            
         return centers, partition_matrix, jm
 
     def _fcm(self, data, n_clusters, m=2, max_iter=1000, error=0.005):
@@ -176,6 +195,72 @@ class Clusterer(object):
         partition_matrix = ut.T
         return centers, partition_matrix, jm
     
+    def _fuzzy_k_protoypes(self, data, categorical_indices, n_clusters, m=2, max_iter=1000, error=0.005):
+        #data: 2d array, size (N, S). N is the number of instances; S is the number of variables
+        #n_clusters: number of clusters
+        #m: fuzzy clustering coefficient
+        #max_it: maximum number of iterations, default=1000
+        #error: stopping criterion, default=0.005
+        #seed: seed for random initialization of u matrix
+        
+        n_instances = data.shape[0]
+        
+        #randomly initaliaze u
+        u = np.random.rand(n_instances, n_clusters)
+        u = np.fmax(u, np.finfo(np.float64).eps)
+        ut = u.T
+        
+        for it in range(0,max_iter):
+            #copy old u matrix
+            u_old = ut.copy()
+            u_old /= np.ones((n_clusters, 1)).dot(np.atleast_2d(u_old.sum(axis=0)))
+            u_old = np.fmax(u_old, np.finfo(np.float64).eps)
+        
+            #elevate to m
+            um = u_old ** m
+        
+            #calculate cluster centers
+            centers = self._FKP_centers(um, data, categorical_indices)
+        
+            #calculate distances
+            dist = self._FKP_distances(centers, data, categorical_indices, numerical_metric='euclidean')
+            dist = np.fmax(dist, np.finfo(np.float64).eps)
+        
+            #calculate objective
+            jm = (um * dist ** 2).sum()
+        
+            #calculate new u matrix
+            ut = dist ** (- 2. / (m - 1))
+            ut /= np.ones((n_clusters, 1)).dot(np.atleast_2d(ut.sum(axis=0)))
+        
+            #stopping criterion
+            if np.linalg.norm(ut - u_old) < error:
+                break
+        
+        partition_matrix = ut.T
+        
+        return centers, partition_matrix, jm
+    
+    def _FKP_centers(self, um, data, categorical_indices):  
+        # If data is categorical find the mode
+        # Take the max of the sum of the columns of the partition matrix
+        
+        # if data is numerical, find the mean
+        centers = um.dot(data) / (np.ones((data.shape[1], 1)).dot(np.atleast_2d(um.sum(axis=1))).T)
+    
+        return centers
+    
+    def _FKP_distances(self, centers, data, categorical_indices, numerical_metric='euclidean'):
+        
+        # If the data is numerical use euclidean distance
+        distances = cdist(centers, data, metric='euclidean')
+        
+        # If the data is categorical, determine if value is different from the mode
+        # if data point == mode, distance = np.finfo(np.float64).eps
+        # if data point ~= mode, distance i= 1
+        
+    
+        return distances
     
     def _fstpso(self, data, n_clusters, max_iter=100, n_particles=None, m=2, path_fit_dump=None, path_sol_dump=None):
         #data: 2d array, size (N, S). N is the number of instances; S is the number of variables
@@ -320,8 +405,10 @@ class Clusterer(object):
         T = 1/(1+tmpt)
 
         return U, T, centers, jm, g
+    
+    ### Gustafson-Kessel
 
-    def _gk(self, m=2, max_iter=100):
+    def _gk(self, m=2, max_iter=1000, error=0.01):
         
         # Initialize the partition matrix
         u = np.random.dirichlet(np.ones(self.data.shape[0]), size=self.nr_clus)
@@ -349,8 +436,10 @@ class Clusterer(object):
             iteration += 1
 
             # Stopping criteria
-            if norm(u - u_old) < self.error:
+            if norm(u - u_old) < error:
                 iteration = max_iter
+                
+        u=np.transpose(u)        
         return centers, u, jm 
 
     def _next_centers_gk(self, data, u, m=2):
@@ -370,7 +459,7 @@ class Clusterer(object):
 
     def _distance_gk(self, data, v, f):
         dif = np.expand_dims(data.reshape(data.shape[0], 1, -1) - v.reshape(1, v.shape[0], -1), axis=3)
-        determ = np.power(np.linalg.det(f), 1 / self.m)
+        determ = np.sign(np.linalg.det(f)) * (np.abs(np.linalg.det(f))) ** (1 / self.m)
         det_time_inv = determ.reshape(-1, 1, 1) * np.linalg.pinv(f)
         temp = np.matmul(dif.transpose((0, 1, 3, 2)), det_time_inv)
         output = np.matmul(temp, dif).squeeze().T
@@ -384,4 +473,97 @@ class Clusterer(object):
         new_u = 1 / new_u.sum(1)
         new_u = new_u.transpose()
         return new_u
+    
+    ### Relational Clustering
+    
+    def _RF_init_centers(self, number_of_clusters, number_of_data_points, relational_data, method='random_initialization'):
+        # Initialize the cluster cenetrs for  relational fuzzy clustering
+        if method == 'random_initialization':       # Use random numbers
+            V = np.random.rand(number_of_clusters,number_of_data_points)
+            V = V/V.sum(axis=1, keepdims=True)
+        elif method == 'randomly_choose_c_rows':    # Use randomly selected data points as initial centers
+            idx = np.random.choice(number_of_data_points,  size=number_of_clusters, replace=False)
+            V = relational_data[idx,:]
+            V = V/V.sum(axis=1, keepdims=True)
+       
+        return V
+
+
+    def _rfcm(self, R, c, m = 2, epsilon = 0.005, maxIter = 1000, initType = 'random_initialization'):
+     
+        #   Relational Fuzzy c-Means (RFCM) for clustering dissimilarity data as
+        #	proposed in [1]. RFCM is the relational dual of Fuzzy c-Means (FCM), 
+        #	so it expects the input relational matrix R to be Euclidean. 
+        #
+        # Output:
+        #               U: fuzzy partition matrix / membership matrix
+        #               V: cluster centers/coefficients
+        #               jm: Fitness function of the best solution
+        #
+        # Input:
+        # R         - the relational (dissimilarity) data matrix of size n x n
+        # c         - number of clusters to be identified
+        # m         - fuzzifier, default 2
+        # epsilon   - convergence criteria, default 0.0001
+        # initType  - initialize relational cluster centers V, default random initialization
+        #               random initialization
+        #               randomly choose c rows from D
+        # maxIter   - the maximum number fo iterations, default 100
+        #
+        # Refs:
+        #   [1] Hathaway, R. J., Davenport, J. W., & Bezdek, J. C. (1989). Relational duals 
+        #   of the c-means clustering algorithms. Pattern recognition, 22(2), 205-212. 
+               
+        # Initialize variables
+        D = np.array(R)         # Relational data
+        n=len(D)                # Number of data points
+        d = np.zeros([c,n])
+        numIter=0
+        stepSize=epsilon
+        
+        # Initialize the membership matrix randomly
+        U = np.random.rand(c, n)
+        U = np.fmax(U, np.finfo(np.float64).eps)
+        
+        # Initialize the (relational) cluster centers
+        V = self._RF_init_centers(number_of_clusters = c, number_of_data_points = n, relational_data = D, method=initType);
+        
+        # Begin the main loop:
+        while numIter<maxIter and stepSize >= epsilon:
+            U0 = U;
+            
+            # Compute the relational distances d between clusters centers V and data points D
+            d=np.zeros([c,n])
+            for i in range(0,c):
+                Vi=V[i,:]
+                tmp1 = D@Vi.T
+                tmp2 = Vi@D@Vi.T/2
+                d[i,:]= tmp1 - tmp2
+                
+            
+            # Update the partition matrix U
+            d = np.power(d,1/(m-1))
+            tmp1 = np.divide(1,d)
+            tmp2 = np.ones([c,1])*sum(tmp1)
+            U = np.divide(np.divide(1,d),tmp2)
+            print(U)
+            
+            # Update cluster centers V
+            V = np.power(U,m)  
+            V = V/V.sum(axis=1, keepdims=True)
+            
+            # Calculate the fitness value
+            jm = (U * d ** 2).sum()
+            
+            # Update the step size
+            stepSize=np.amax(abs(U-U0))
+
+            if self._verbose == True:
+                print('pyFUME just finished iteration '+ str(numIter) + ' of the RFCM clustering algoritm. The (maximum) difference between the previous and the newly computed partition matrix is' + str(stepSize) + '.')
+            
+            numIter = numIter + 1;
+        
+        # Return the cluster center location, the membership matrix and the fitness value
+        return V, U, jm
+        
 
