@@ -81,7 +81,11 @@ class DataLoader(object):
         self.dataY=self.data[:,-1]
 
         if normalize=='minmax' or normalize=='linear' or normalize==True:
-            self.dataX = (self.dataX - np.abs(self.dataX).min(axis=0)) / (np.abs(self.dataX).max(axis=0)-np.abs(self.dataX).min(axis=0))
+            mini=np.abs(self.dataX).min(axis=0)
+            maxi=np.abs(self.dataX).max(axis=0)
+            self.normalization_values=zip(self.variable_names, mini, maxi)
+
+            self.dataX = (self.dataX - self.dataX.min(axis=0) / self.dataX.max(axis=0)-self.dataX.min(axis=0))
 
         elif normalize == 'zscore':
             self.dataX = (self.dataX - self.dataX.mean(axis=0)) / self.dataX.std(axis=0)
@@ -92,7 +96,20 @@ class DataLoader(object):
         num_cols = DF._get_numeric_data().columns
         categ_cols = list( set(cols)-set(num_cols))
         return categ_cols
-
+    
+    def get_input_data(self):
+        return self.dataX
+    
+    def get_target_data(self):
+        return self.dataY
+    def get_all_data(self):
+        return self.data
+    
+    def get_variable_names(self):
+        return self.variable_names
+    
+    def get_normalization_values(self):
+        return self.normalization_values
 
 
 if __name__ == "__main__":
